/* tc-or1.c -- Assemble for the OpenRISC 1000.
   Copyright (C) 1989, 1990, 1991, 1992, 1993 Free Software Foundation, Inc.
   Contributed by Damjan Lampret <lampret@opencores.org>.
   Based upon a29k port.

   This file is part of GAS, the GNU Assembler.

   GAS is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GAS is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GAS; see the file COPYING.  If not, write to
   the Free Software Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/* tc-a29k.c used as a template.  */

#include <ctype.h>
#include "as.h"

#include "opcode/or1.h"

#define DEBUG 0

/* Make it easier to clone this machine desc into another one.  */
#define	machine_opcode	or1_opcode
#define	machine_opcodes	or1_opcodes
#define	machine_ip	or1_ip
#define	machine_it	or1_it

/* handle of the OPCODE hash table */
static struct hash_control *op_hash = NULL;

struct machine_it
  {
    char *error;
    unsigned long opcode;
    struct nlist *nlistp;
    expressionS exp;
    int pcrel;
    int reloc_offset;		/* Offset of reloc within insn */

    int reloc;
  }
the_insn;

static void machine_ip PARAMS ((char *str));
/* static void print_insn PARAMS ((struct machine_it *insn)); */
#ifndef OBJ_COFF
static void s_use PARAMS ((int));
#endif

const pseudo_typeS
md_pseudo_table[] =
{
  {"align", s_align_bytes, 4},
  {"block", s_space, 0},
  {"cputype", s_ignore, 0},	/* CPU as 29000 or 29050 */
  {"reg", s_lsym, 0},		/* Register equate, same as equ */
  {"space", s_ignore, 0},	/* Listing control */
  {"sect", s_ignore, 0},	/* Creation of coff sections */
  {"proc", s_ignore, 0},	/* Start of a function */
  {"endproc", s_ignore, 0},	/* Function end */
#ifndef OBJ_COFF
  /* We can do this right with coff.  */
  {"use", s_use, 0},
#endif
  {"word", cons, 4},
  {NULL, 0, 0},
};

int md_short_jump_size = 4;
int md_long_jump_size = 4;
#if defined(BFD_HEADERS)
#ifdef RELSZ
const int md_reloc_size = RELSZ;	/* Coff headers */
#else
const int md_reloc_size = 12;		/* something else headers */
#endif
#else
const int md_reloc_size = 12;		/* Not bfdized*/
#endif

/* This array holds the chars that always start a comment.  If the
   pre-processor is disabled, these aren't very useful */
const char comment_chars[] = ";";

/* This array holds the chars that only start a comment at the beginning of
   a line.  If the line seems to have the form '# 123 filename'
   .line and .file directives will appear in the pre-processed output */
/* Note that input_file.c hand checks for '#' at the beginning of the
   first line of the input file.  This is because the compiler outputs
   #NO_APP at the beginning of its output. */
/* Also note that comments like this one will always work */
const char line_comment_chars[] = "#";

/* We needed an unused char for line separation to work around the
   lack of macros, using sed and such.  */
const char line_separator_chars[] = "@";

/* Chars that can be used to separate mant from exp in floating point nums */
const char EXP_CHARS[] = "eE";

/* Chars that mean this number is a floating point constant */
/* As in 0f12.456 */
/* or    0d1.2345e12 */
const char FLT_CHARS[] = "rRsSfFdDxXpP";

/* Also be aware that MAXIMUM_NUMBER_OF_CHARS_FOR_FLOAT may have to be
   changed in read.c.  Ideally it shouldn't have to know about it at
   all, but nothing is ideal around here.  */

#ifndef OBJ_COFF

static void
s_use (ignore)
     int ignore;
{
  if (strncmp (input_line_pointer, ".text", 5) == 0)
    {
      input_line_pointer += 5;
      s_text (0);
      return;
    }
  if (strncmp (input_line_pointer, ".data", 5) == 0)
    {
      input_line_pointer += 5;
      s_data (0);
      return;
    }
  /* Literals can't go in the text segment because you can't read from
     instruction memory on some 29k's.  So, into initialized data. */
  if (strncmp (input_line_pointer, ".lit", 4) == 0)
    {
      input_line_pointer += 4;
      subseg_set (SEG_DATA, 200);
      demand_empty_rest_of_line ();
      return;
    }

  as_bad ("Unknown segment type");
  demand_empty_rest_of_line ();
}

#endif /* OBJ_COFF */

/* Install symbol definition that maps REGNAME to REGNO.
   FIXME-SOON:  These are not recognized in mixed case.  */

static void
insert_sreg (regname, regnum)
     char *regname;
     int regnum;
{
  /* FIXME-SOON, put something in these syms so they won't be output
     to the symbol table of the resulting object file.  */

  /* Must be large enough to hold the names of the special registers.  */
  char buf[80];
  int i;

  symbol_table_insert (symbol_new (regname, SEG_REGISTER, (valueT) regnum,
				   &zero_address_frag));
  for (i = 0; regname[i]; i++)
    buf[i] = islower (regname[i]) ? toupper (regname[i]) : regname[i];
  buf[i] = '\0';

  symbol_table_insert (symbol_new (buf, SEG_REGISTER, (valueT) regnum,
				   &zero_address_frag));
}

/* Install symbol definitions for assorted special registers.
   See ASM29K Ref page 2-9.  */

void
define_some_regs ()
{
  insert_sreg ("mmu", 1);
  insert_sreg ("r5", 5);
}

/* This function is called once, at assembler startup time.  It should
   set up all the tables, etc., that the MD part of the assembler will
   need.  */
void
md_begin ()
{
  register const char *retval = NULL;
  int lose = 0;
  register int skipnext = 0;
  register unsigned int i;
  register char *strend, *strend2;

  /* Hash up all the opcodes for fast use later.  */

  op_hash = hash_new ();

  for (i = 0; i < num_opcodes; i++)
    {
      const char *name = machine_opcodes[i].name;

      if (skipnext)
	{
	  skipnext = 0;
	  continue;
	}

      retval = hash_insert (op_hash, name, (PTR) &machine_opcodes[i]);
      if (retval != NULL)
	{
	  fprintf (stderr, "internal error: can't hash `%s': %s\n",
		   machine_opcodes[i].name, retval);
	  lose = 1;
	}
    }

  if (lose)
    as_fatal ("Broken assembler.  No assembly attempted.");

  define_some_regs ();
}

/* Set bits in machine opcode according to insn->encoding
   description and passed operand. */
   
void encode(struct machine_opcode *insn, unsigned long *opcode, signed long param_val, char param_ch)
{
	int opc_pos = 0;
	int param_pos = 0;
	char *enc;

#if DEBUG
	printf("    encode:  opcode=%.8lx  param_val=%.8lx abs=%.8lx param_ch=%c\n", *opcode, param_val, abs(param_val), param_ch);
#endif
	for (enc = insn->encoding; *enc != '\0'; enc++)
		if (*enc == param_ch)
			param_pos++;
#if 0			
	if (param_val >> param_pos)
		as_bad ("Immediate too large for `%s' instruction.", insn->name);
#endif

	opc_pos = 32;
	for (enc = insn->encoding; *enc != '\0'; )
		if ((*enc == '0') && (*(enc+1) == 'x')) {
			int tmp = strtol(enc, NULL, 16);
			opc_pos -= 4;
			*opcode |= tmp << opc_pos;
			enc += 3;
		}
		else if ((*enc == '0') || (*enc == '-')) {
			opc_pos--;
			enc++;
		}
		else if (*enc == '1') {
			opc_pos--;
			*opcode |= 1 << opc_pos;
			enc++;
		}
		else if (*enc == param_ch) {
			opc_pos--;
			param_pos--;
			*opcode |= ((param_val >> param_pos) & 0x1) << opc_pos;
			enc++;
		}
                else if (isalpha(*enc)) {
			opc_pos--;
                        enc++;
		}
		else
			enc++;
	
}

/* Assemble a single instruction.  Its label has already been handled
   by the generic front end.  We just parse opcode and operands, and
   produce the bytes of data and relocation.  */

void
md_assemble (str)
     char *str;
{
  char *toP;

#if DEBUG
  printf("NEW INSTRUCTION\n");
#endif
  know (str);
  machine_ip (str);
  toP = frag_more (4);
  /* put out the opcode */
  md_number_to_chars (toP, the_insn.opcode, 4);
#if DEBUG
  print_insn(&the_insn); 
#endif

  /* put out the symbol-dependent stuff */
  if (the_insn.reloc != NO_RELOC)
    {
      fix_new_exp (frag_now,
		   (toP - frag_now->fr_literal + the_insn.reloc_offset),
		   4,		/* size */
		   &the_insn.exp,
		   the_insn.pcrel,
		   the_insn.reloc);
    }
}

char *
parse_operand (s, operandp, opt)
     char *s;
     expressionS *operandp;
     int opt;
{
  char *save = input_line_pointer;
  char *new;

#if DEBUG
  printf("  PROCESS NEW OPERAND\n");
#endif
  input_line_pointer = s;
  expression (operandp);
  if (operandp->X_op == O_absent && ! opt)
    as_bad ("missing operand");
  new = input_line_pointer;
  input_line_pointer = save;
  
  if ((operandp->X_op == O_symbol) && (*s != '_')) {
	for(save = s; s < new; s++)
		if ((*s == 'r') && isdigit(*(s+1))) {
	  		operandp->X_add_number = strtol(s+1, NULL, 10);
	  		operandp->X_op = O_register;
	  	}
	s = save;
  }

#if DEBUG
  printf("  parse_operand(%s): operandp->X_op = %u\n", s, operandp->X_op);
#endif

  return new;
}

/* Instruction parsing.  Takes a string containing the opcode.
   Operands are at input_line_pointer.  Output is in the_insn.
   Warnings or errors are generated.  */

static void
machine_ip (str)
     char *str;
{
  char *s;
  const char *args;
  struct machine_opcode *insn;
  char *argsStart;
  unsigned long opcode;
  expressionS the_operand;
  expressionS *operand = &the_operand;
  unsigned int regno;
  int reloc = NO_RELOC;

#if DEBUG
  printf("machine_ip(%s)\n", str);
#endif

  s = str;
  for (; isalnum (*s) || *s == '.'; ++s)
    if (isupper (*s))
	*s = tolower (*s);

  switch (*s)
    {
    case '\0':
      break;

    case ' ':			/* FIXME-SOMEDAY more whitespace */
      *s++ = '\0';
      break;

    default:
      as_bad ("Unknown opcode1: `%s'", str);
      return;
    }
  if ((insn = (struct machine_opcode *) hash_find (op_hash, str)) == NULL)
    {
      as_bad ("Unknown opcode2 `%s'.", str);
      return;
    }
  argsStart = s;
  opcode = 0;
  memset (&the_insn, '\0', sizeof (the_insn));
  the_insn.reloc = NO_RELOC;

  /* Build the opcode, checking as we go to make sure that the
     operands match.
   
     If an operand matches, we modify the_insn or opcode appropriately,
     and do a "continue".  If an operand fails to match, we "break".  */

  if (insn->args[0] != '\0')
    {
      /* Prime the pump.  */
      s = parse_operand (s, operand, insn->args[0] == 'I');
    }

  for (args = insn->args;; ++args)
    {
#if DEBUG
      printf("  args = %s\n", args);
#endif
      switch (*args)
	{

	case '\0':		/* end of args */
	  if (*s == '\0')
	    {
	      /* We are truly done. */
	      the_insn.opcode = opcode;
	      return;
	    }
	  as_bad ("Too many operands: %s", s);
	  break;

	case ',':		/* Must match a comma */
	  if (*s++ == ',')
	    {
	      reloc = NO_RELOC;
	      /* Parse next operand.  */
	      s = parse_operand (s, operand, args[1] == 'I');
	      continue;
	    }
	  break;

	case '(':		/* Must match a ( */
	    s = parse_operand (s, operand, args[1] == 'I');
          continue;
	  
	case ')':		/* Must match a ) */
	  continue;

	case 'r':		/* A general register */
	  args++;
	  if (operand->X_op != O_register)
	  	break;		/* Only registers */
	  know (operand->X_add_symbol == 0);
	  know (operand->X_op_symbol == 0);
	  regno = operand->X_add_number;
          encode(insn, &opcode, regno, *args);
          printf("    r: operand->X_op = %d\n", operand->X_op);
	  continue;

	default:
	  /* if (!isalpha(*args))
	  	break;	*/	/* Only immediate values */
	  
	  
	  if (strncmp(args, "lo(", 3) == 0) {
	  	printf("reloc_const\n");
	  	reloc = RELOC_CONST;
	  }
	  else if (strncmp(args, "hi(", 3) == 0) {
	  	printf("reloc_consth\n");
	  	reloc = RELOC_CONSTH;
	  }
	  	  	
	  if (*s == '(') {
	  	operand->X_op = O_constant;
	  	operand->X_add_number = 0;
	  }
	  printf("    default case: operand->X_op = %d, *args = %s, *s = %s\n", operand->X_op, args, s);
	  
	  if (operand->X_op == O_constant)
	    {
	      {
		bfd_vma v, mask;
		mask = 0x1ffff;
		v = abs(operand->X_add_number) & ~ mask;
		printf("    v = %d\n", v);
		if (v)
		  as_bad ("call/jmp target out of range (1)");
	      }
	      encode(insn, &opcode, operand->X_add_number, *args);
	      the_insn.reloc = reloc;
              printf("    reloc const=%d\n", the_insn.reloc);
	      continue;
	    }
	  if (reloc == NO_RELOC)
	  	the_insn.reloc = RELOC_JUMPTARG;
	  else
	  	the_insn.reloc = reloc;
         /*  the_insn.reloc = insn->reloc;  */
          printf("    reloc sym=%d\n", the_insn.reloc);
	  the_insn.exp = *operand;
	 /*  the_insn.reloc_offset = 1; */
	  the_insn.pcrel = 1;	/* Assume PC-relative jump */
	  /* FIXME-SOON, Do we figure out whether abs later, after
             know sym val? */
          encode(insn, &opcode, operand->X_add_number, *args);
	  continue;
	}
	
      /* Types or values of args don't match.  */
      as_bad ("Invalid operands");
      return;
    }
}

/* This is identical to the md_atof in m68k.c.  I think this is right,
   but I'm not sure.

   Turn a string in input_line_pointer into a floating point constant
   of type type, and store the appropriate bytes in *litP.  The number
   of LITTLENUMS emitted is stored in *sizeP .  An error message is
   returned, or NULL on OK.  */

/* Equal to MAX_PRECISION in atof-ieee.c */
#define MAX_LITTLENUMS 6

char *
md_atof (type, litP, sizeP)
     char type;
     char *litP;
     int *sizeP;
{
  int prec;
  LITTLENUM_TYPE words[MAX_LITTLENUMS];
  LITTLENUM_TYPE *wordP;
  char *t;

  switch (type)
    {

    case 'f':
    case 'F':
    case 's':
    case 'S':
      prec = 2;
      break;

    case 'd':
    case 'D':
    case 'r':
    case 'R':
      prec = 4;
      break;

    case 'x':
    case 'X':
      prec = 6;
      break;

    case 'p':
    case 'P':
      prec = 6;
      break;

    default:
      *sizeP = 0;
      return "Bad call to MD_ATOF()";
    }
  t = atof_ieee (input_line_pointer, type, words);
  if (t)
    input_line_pointer = t;
  *sizeP = prec * sizeof (LITTLENUM_TYPE);
  for (wordP = words; prec--;)
    {
      md_number_to_chars (litP, (valueT) (*wordP++), sizeof (LITTLENUM_TYPE));
      litP += sizeof (LITTLENUM_TYPE);
    }
  return 0;
}

/*
 * Write out big-endian.
 */
void
md_number_to_chars (buf, val, n)
     char *buf;
     valueT val;
     int n;
{
  number_to_chars_bigendian (buf, val, n);
}

void
md_apply_fix (fixP, val)
     fixS *fixP;
     long val;
{
  char *buf = fixP->fx_where + fixP->fx_frag->fr_literal;

  printf("md_apply_fix val:%x\n",val);

  fixP->fx_addnumber = val;	/* Remember value for emit_reloc */


  know (fixP->fx_size == 4);
  know (fixP->fx_r_type < NO_RELOC);

  /* This is a hack.  There should be a better way to handle this.  */
  if (fixP->fx_r_type == RELOC_WDISP30 && fixP->fx_addsy)
    {
      val += fixP->fx_where + fixP->fx_frag->fr_address;
    }

  switch (fixP->fx_r_type)
    {

    case RELOC_32:
      buf[0] = val >> 24;
      buf[1] = val >> 16;
      buf[2] = val >> 8;
      buf[3] = val;
      break;

    case RELOC_8:
      buf[0] = val;
      break;

    case RELOC_WDISP30:
      val = (val >>= 2) + 1;
      buf[0] |= (val >> 24) & 0x3f;
      buf[1] = (val >> 16);
      buf[2] = val >> 8;
      buf[3] = val;
      break;

    case RELOC_HI22:
      buf[1] |= (val >> 26) & 0x3f;
      buf[2] = val >> 18;
      buf[3] = val >> 10;
      break;

    case RELOC_LO10:
      buf[2] |= (val >> 8) & 0x03;
      buf[3] = val;
      break;

    case RELOC_BASE13:
      buf[2] |= (val >> 8) & 0x1f;
      buf[3] = val;
      break;

    case RELOC_WDISP22:
      val = (val >>= 2) + 1;
      /* FALLTHROUGH */
    case RELOC_BASE22:
      buf[1] |= (val >> 16) & 0x3f;
      buf[2] = val >> 8;
      buf[3] = val;
      break;

    case RELOC_JUMPTARG:	/* 00XX00XX pattern in a word */
      if (!fixP->fx_done)
	{
	  /* The linker tries to support both AMD and old GNU style
             R_IREL relocs.  That means that if the addend is exactly
             the negative of the address within the section, the
             linker will not handle it correctly.  */
#if 0
	  if (fixP->fx_pcrel
	      && val != 0
	      && val == - (fixP->fx_frag->fr_address + fixP->fx_where))
	    as_bad_where
	      (fixP->fx_file, fixP->fx_line,
	       "the linker will not handle this relocation correctly (1)");
#endif
	}
      else if (fixP->fx_pcrel)
	{
	  long v = val >> 17;
	  if (v != 0 && v != -1)
	    as_bad_where (fixP->fx_file, fixP->fx_line,
			  "call/jmp target out of range (2)");
	}
      else
	/* this case was supposed to be handled in machine_ip */
	abort ();
      buf[2] = val >> 10;	/* Holds bits 0003FFFC of address */
      buf[3] = val >> 2;
      break;

    case RELOC_CONST:		/* 0000XXXX pattern in a word */
      buf[2] = val >> 8;	/* Holds bits 0000XXXX */
      buf[3] = val;
      break;

    case RELOC_CONSTH:		/* 0000XXXX pattern in a word */
      buf[2] = val >> 24;	/* Holds bits XXXX0000 */
      buf[3] = val >> 16;
      break;

    case NO_RELOC:
    default:
      as_bad ("bad relocation type: 0x%02x", fixP->fx_r_type);
      break;
    }
}

#ifdef OBJ_COFF
short
tc_coff_fix2rtype (fixP)
     fixS *fixP;
{

  printf("tc_coff_fix2rtype\n");
  
  switch (fixP->fx_r_type)
    {
    case RELOC_32:
      return (R_WORD);
    case RELOC_8:
      return (R_BYTE);
    case RELOC_CONST:
      return (R_ILOHALF);
    case RELOC_CONSTH:
      return (R_IHIHALF);
    case RELOC_JUMPTARG:
      return (R_IREL);
    default:
      printf ("need %o3\n", fixP->fx_r_type);
      abort ();
    }				/* switch on type */

  return (0);
}

#endif /* OBJ_COFF */

/* should never be called for 29k */
void
md_create_short_jump (ptr, from_addr, to_addr, frag, to_symbol)
     char *ptr;
     addressT from_addr, to_addr;
     fragS *frag;
     symbolS *to_symbol;
{
  as_fatal ("or1_create_short_jmp\n");
}

/* should never be called for 29k */
void
md_convert_frag (headers, seg, fragP)
     object_headers *headers;
     segT seg;
     register fragS *fragP;
{
  as_fatal ("or1_convert_frag\n");
}

/* should never be called for 29k */
void
md_create_long_jump (ptr, from_addr, to_addr, frag, to_symbol)
     char *ptr;
     addressT from_addr;
     addressT to_addr;
     fragS *frag;
     symbolS *to_symbol;
{
  as_fatal ("or1_create_long_jump\n");
}

/* should never be called for or1 */
int
md_estimate_size_before_relax (fragP, segtype)
     register fragS *fragP;
     segT segtype;
{
  as_fatal ("or1_estimate_size_before_relax\n");
  return 0;
}

#if DEBUG
/* for debugging only */
static void
print_insn (insn)
     struct machine_it *insn;
{
  char *Reloc[] =
  {
    "RELOC_8",
    "RELOC_16",
    "RELOC_32",
    "RELOC_DISP8",
    "RELOC_DISP16",
    "RELOC_DISP32",
    "RELOC_WDISP30",
    "RELOC_WDISP22",
    "RELOC_HI22",
    "RELOC_22",
    "RELOC_13",
    "RELOC_LO10",
    "RELOC_SFA_BASE",
    "RELOC_SFA_OFF13",
    "RELOC_BASE10",
    "RELOC_BASE13",
    "RELOC_BASE22",
    "RELOC_PC10",
    "RELOC_PC22",
    "RELOC_JMP_TBL",
    "RELOC_SEGOFF16",
    "RELOC_GLOB_DAT",
    "RELOC_JMP_SLOT",
    "RELOC_RELATIVE",
    "NO_RELOC"
  };

  if (insn->error)
    {
      fprintf (stderr, "ERROR: %s\n");
    }
  fprintf (stderr, "reloc = %s\n", Reloc[insn->reloc]);
  fprintf (stderr, "exp =  {\n");
  fprintf (stderr, "\t\tX_add_symbol = %s\n",
	   insn->exp.X_add_symbol ?
	   (S_GET_NAME (insn->exp.X_add_symbol) ?
	    S_GET_NAME (insn->exp.X_add_symbol) : "???") : "0");
  fprintf (stderr, "\t\tX_op_symbol = %s\n",
	   insn->exp.X_op_symbol ?
	   (S_GET_NAME (insn->exp.X_op_symbol) ?
	    S_GET_NAME (insn->exp.X_op_symbol) : "???") : "0");
  fprintf (stderr, "\t\tX_add_number = %d\n",
	   insn->exp.X_add_number);
  fprintf (stderr, "}\n");
}

#endif

/* Translate internal representation of relocation info to target format.

   On sparc/29k: first 4 bytes are normal unsigned long address, next three
   bytes are index, most sig. byte first.  Byte 7 is broken up with
   bit 7 as external, bits 6 & 5 unused, and the lower
   five bits as relocation type.  Next 4 bytes are long addend. */
/* Thanx and a tip of the hat to Michael Bloom, mb@ttidca.tti.com */

#ifdef OBJ_AOUT

void
tc_aout_fix_to_chars (where, fixP, segment_address_in_file)
     char *where;
     fixS *fixP;
     relax_addressT segment_address_in_file;
{
  long r_symbolnum;

  printf("tc_aout_fix_to_chars\n");
  
  know (fixP->fx_r_type < NO_RELOC);
  know (fixP->fx_addsy != NULL);

  md_number_to_chars (where,
       fixP->fx_frag->fr_address + fixP->fx_where - segment_address_in_file,
		      4);

  r_symbolnum = (S_IS_DEFINED (fixP->fx_addsy)
		 ? S_GET_TYPE (fixP->fx_addsy)
		 : fixP->fx_addsy->sy_number);

  where[4] = (r_symbolnum >> 16) & 0x0ff;
  where[5] = (r_symbolnum >> 8) & 0x0ff;
  where[6] = r_symbolnum & 0x0ff;
  where[7] = (((!S_IS_DEFINED (fixP->fx_addsy)) << 7) & 0x80) | (0 & 0x60) | (fixP->fx_r_type & 0x1F);
  /* Also easy */
  md_number_to_chars (&where[8], fixP->fx_addnumber, 4);
}

#endif /* OBJ_AOUT */

CONST char *md_shortopts = "";
struct option md_longopts[] = {
  {NULL, no_argument, NULL, 0}
};
size_t md_longopts_size = sizeof(md_longopts);

int
md_parse_option (c, arg)
     int c;
     char *arg;
{
  return 0;
}

void
md_show_usage (stream)
     FILE *stream;
{
}

/* This is called when a line is unrecognized.  This is used to handle
   definitions of or1 style local labels.  */

int
or1_unrecognized_line (c)
     int c;
{
  int lab;
  char *s;

  if (c != '$'
      || ! isdigit ((unsigned char) input_line_pointer[0]))
    return 0;

  s = input_line_pointer;

  lab = 0;
  while (isdigit ((unsigned char) *s))
    {
      lab = lab * 10 + *s - '0';
      ++s;
    }

  if (*s != ':')
    {
      /* Not a label definition.  */
      return 0;
    }

  if (dollar_label_defined (lab))
    {
      as_bad ("label \"$%d\" redefined", lab);
      return 0;
    }

  define_dollar_label (lab);
  colon (dollar_label_name (lab, 0));
  input_line_pointer = s + 1;

  return 1;
}

/* Default the values of symbols known that should be "predefined".  We
   don't bother to predefine them unless you actually use one, since there
   are a lot of them.  */

symbolS *
md_undefined_symbol (name)
     char *name;
{
  long regnum;
  char testbuf[5 + /*SLOP*/ 5];

  printf("md_undefined_symbol(%s)\n", name);
  /* Register name */
  if (name[0] == 'r' || name[0] == 'R' || name[0] == 'a' || name[0] == 'b') {
	long maxreg;

	  /* Parse the number, make sure it has no extra zeroes or
	     trailing chars. */
	  regnum = atol (&name[1]);

	  if (regnum > 31)
	    as_fatal("Register out of range");

	  sprintf (testbuf, "%ld", regnum);
	  if (strcmp (testbuf, &name[1]) != 0)
	    return NULL;	/* gr007 or lr7foo or whatever */

	  /* We have a wiener!  Define and return a new symbol for it.  */
	  return (symbol_new (name, SEG_REGISTER, (valueT) regnum,
			      &zero_address_frag));
  }

  return NULL;
}

/* Parse an operand that is machine-specific.  */

void
md_operand (expressionP)
     expressionS *expressionP;
{
#if DEBUG
  printf("  md_operand(input_line_pointer = %s)\n", input_line_pointer);
#endif
  if (input_line_pointer[0] == '%' && input_line_pointer[1] == 'r')
    {
      /* We have a numeric register expression.  No biggy.  */
      input_line_pointer += 2;	/* Skip %r */
      (void) expression (expressionP);
      if (expressionP->X_op != O_constant
	  || expressionP->X_add_number > 255)
	as_bad ("Invalid expression after %%%%\n");
      expressionP->X_op = O_register;
    }
  else if (input_line_pointer[0] == '&')
    {
      /* We are taking the 'address' of a register...this one is not
	 in the manual, but it *is* in traps/fpsymbol.h!  What they
	 seem to want is the register number, as an absolute number.  */
      input_line_pointer++;	/* Skip & */
      (void) expression (expressionP);
      if (expressionP->X_op != O_register)
	as_bad ("Invalid register in & expression");
      else
	expressionP->X_op = O_constant;
    }
  else if (input_line_pointer[0] == '$'
	   && isdigit ((unsigned char) input_line_pointer[1]))
    {
      long lab;
      char *name;
      symbolS *sym;

      /* This is a local label.  */
      ++input_line_pointer;
      lab = (long) get_absolute_expression ();
      if (dollar_label_defined (lab))
	{
	  name = dollar_label_name (lab, 0);
	  sym = symbol_find (name);
	}
      else
	{
	  name = dollar_label_name (lab, 1);
	  sym = symbol_find_or_make (name);
	}

      expressionP->X_op = O_symbol;
      expressionP->X_add_symbol = sym;
      expressionP->X_add_number = 0;
    }
  else if (input_line_pointer[0] == '$')
    {
      char *s;
      char type;
      int fieldnum, fieldlimit;
      LITTLENUM_TYPE floatbuf[8];

      /* $float(), $doubleN(), or $extendN() convert floating values
	 to integers.  */

      s = input_line_pointer;

      ++s;

      fieldnum = 0;
      if (strncmp (s, "double", sizeof "double" - 1) == 0)
	{
	  s += sizeof "double" - 1;
	  type = 'd';
	  fieldlimit = 2;
	}
      else if (strncmp (s, "float", sizeof "float" - 1) == 0)
	{
	  s += sizeof "float" - 1;
	  type = 'f';
	  fieldlimit = 1;
	}
      else if (strncmp (s, "extend", sizeof "extend" - 1) == 0)
	{
	  s += sizeof "extend" - 1;
	  type = 'x';
	  fieldlimit = 4;
	}
      else 
	{
	  return;
	}

      if (isdigit (*s))
	{
	  fieldnum = *s - '0';
	  ++s;
	}
      if (fieldnum >= fieldlimit)
	return;

      SKIP_WHITESPACE ();
      if (*s != '(')
	return;
      ++s;
      SKIP_WHITESPACE ();

      s = atof_ieee (s, type, floatbuf);
      if (s == NULL)
	return;
      s = s;

      SKIP_WHITESPACE ();
      if (*s != ')')
	return;
      ++s;
      SKIP_WHITESPACE ();

      input_line_pointer = s;
      expressionP->X_op = O_constant; 
      expressionP->X_unsigned = 1;
      expressionP->X_add_number = ((floatbuf[fieldnum * 2]
				    << LITTLENUM_NUMBER_OF_BITS)
				   + floatbuf[fieldnum * 2 + 1]);
    }
}

/* Round up a section size to the appropriate boundary.  */
valueT
md_section_align (segment, size)
     segT segment;
     valueT size;
{
  return size;			/* Byte alignment is fine */
}

/* Exactly what point is a PC-relative offset relative TO?
   On the 29000, they're relative to the address of the instruction,
   which we have set up as the address of the fixup too.  */
long
md_pcrel_from (fixP)
     fixS *fixP;
{
  return fixP->fx_where + fixP->fx_frag->fr_address;
}

/* end of tc-or1.c */
