/* Table of opcodes for the OpenRISC 1001 implementation (OR1K family).
   Copyright 1990, 1991, 1992, 1993 Free Software Foundation, Inc.
   Contributed by Damjan Lampret (lampret@opencores.org).
   
   This is obsolete (at least at the moment).

This file is part of GDB and GAS.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

struct or1_opcode {
  /* Name of the instruction.  */
  char *name;

  /* A string of characters which describe the operands.
     Valid characters are:
     ,	 Itself.  The character appears in the assembly code.
     rA	 Register operand.
     rB  Register operand.
     rD  Register operand.
     I	 An immediate operand, range -32768 to 32767.
     J	 An immediate operand, range -65536 to 65535.
     K	 An immediate operand, range 0 to 65535.
     L	 An immediate operand, range 0 to 31.
     M	 An immediate operand, range -128 to 127.
     N	 An immediate operand, range -8 to 7.
     O	 An immediate operand, range  to .  */

  char *args;
  
  /* Opcode and operand encoding. */
  char *encoding;
};

#ifndef CONST
#define CONST
#endif /* CONST */

/* Opcode encoding:
   opcode[31:30]: 00 - neither of operands is register
   		  01 - one destination (rD) and one source (rA) register
   		  10 - two source registers (rA, rB), no destination register
   		  11 - all three operands are registers
   opcode[29:28]: instruction type
   opcode[27:XX]: instruction parameters (specific to individual insn type)
*/

static CONST struct or1_opcode or1_opcodes[] =
{

{ "nop",     "",             "0x0 0000 0000 0000 0000 0000 0000 0000", },
{ "jal",     "X",            "0x1 0XXX XXXX XXXX XXXX XXXX XXXX XXXX", },
{ "bal",     "X",            "0x2 0XXX XXXX XXXX XXXX XXXX XXXX XXXX", },
{ "bnf",     "X",            "0x3 0XXX XXXX XXXX XXXX XXXX XXXX XXXX", },
{ "bf",      "X",            "0x3 1XXX XXXX XXXX XXXX XXXX XXXX XXXX", },

{ "lw",      "rD,I(rA)",     "0x4 000D DDDD AAAA IIII IIII IIII IIII", },
{ "lbz",     "rD,I(rA)",     "0x4 010D DDDD AAAA IIII IIII IIII IIII", },
{ "lbs",     "rD,I(rA)",     "0x4 011D DDDD AAAA IIII IIII IIII IIII", },
{ "lhz",     "rD,I(rA)",     "0x4 100D DDDD AAAA IIII IIII IIII IIII", },
{ "lhs",     "rD,I(rA)",     "0x4 101D DDDD AAAA IIII IIII IIII IIII", },

{ "mtsr",    "rS,rA",        "0x4 0000 0000 AAAA 0000 SSSS SSSS SSSS", },
{ "mfsr",    "rA,rS",        "0x4 0000 0000 AAAA 0000 SSSS SSSS SSSS", },

{ "addi",    "rD,rA,I",      "0x5 000D DDDD AAAA IIII IIII IIII IIII", },
{ "subi",    "rD,rA,I",      "0x5 001D DDDD AAAA IIII IIII IIII IIII", },
{ "andi",    "rD,rA,I",      "0x5 010D DDDD AAAA IIII IIII IIII IIII", },
{ "ori",     "rD,rA,I",      "0x5 100D DDDD AAAA IIII IIII IIII IIII", },
{ "xori",    "rD,rA,I",      "0x5 110D DDDD AAAA IIII IIII IIII IIII", },
{ "muli",    "rD,rA,I",      "0x6 000D DDDD AAAA IIII IIII IIII IIII", },
{ "slli",    "rD,rA,I",      "0x7 000D DDDD AAAA ---- ---- ---I IIII", },
{ "srli",    "rD,rA,I",      "0x7 100D DDDD AAAA ---- ---- ---I IIII", },
{ "srai",    "rD,rA,I",      "0x7 101D DDDD AAAA ---- ---- ---I IIII", },

{ "sw",      "J(rA),rB",     "0x8 0000 ---- AAAA BBBB JJJJ JJJJ JJJJ", },
{ "sb",      "J(rA),rB",     "0x8 0100 ---- AAAA BBBB JJJJ JJJJ JJJJ", },
{ "sh",      "J(rA),rB",     "0x8 1000 ---- AAAA BBBB JJJJ JJJJ JJJJ", },

{ "sfeq",    "rA,rB",        "0x9 0000 ---- AAAA BBBB ---- ---- ----", },
{ "sfne",    "rA,rB",        "0x9 0001 ---- AAAA BBBB ---- ---- ----", },
{ "sfgts",   "rA,rB",        "0x9 0010 ---- AAAA BBBB ---- ---- ----", },
{ "sfges",   "rA,rB",        "0x9 0011 ---- AAAA BBBB ---- ---- ----", },
{ "sflts",   "rA,rB",        "0x9 0100 ---- AAAA BBBB ---- ---- ----", },
{ "sfles",   "rA,rB",        "0x9 0101 ---- AAAA BBBB ---- ---- ----", },
{ "sfgtu",   "rA,rB",        "0x9 0110 ---- AAAA BBBB ---- ---- ----", },
{ "sfgeu",   "rA,rB",        "0x9 0111 ---- AAAA BBBB ---- ---- ----", },
{ "sfltu",   "rA,rB",        "0x9 1000 ---- AAAA BBBB ---- ---- ----", },
{ "sfleu",   "rA,rB",        "0x9 1001 ---- AAAA BBBB ---- ---- ----", },

{ "sfeqi",   "rA,I",         "0xA 0000 ---- AAAA IIII IIII IIII IIII", },
{ "sfnei",   "rA,I",         "0xA 0001 ---- AAAA IIII IIII IIII IIII", },
{ "sfgtsi",  "rA,I",         "0xA 0010 ---- AAAA IIII IIII IIII IIII", },
{ "sfgesi",  "rA,I",         "0xA 0011 ---- AAAA IIII IIII IIII IIII", },
{ "sfltsi",  "rA,I",         "0xA 0100 ---- AAAA IIII IIII IIII IIII", },
{ "sflesi",  "rA,I",         "0xA 0101 ---- AAAA IIII IIII IIII IIII", },
{ "sfgtui",  "rA,I",         "0xA 0110 ---- AAAA IIII IIII IIII IIII", },
{ "sfgeui",  "rA,I",         "0xA 0111 ---- AAAA IIII IIII IIII IIII", },
{ "sfltui",  "rA,I",         "0xA 1000 ---- AAAA IIII IIII IIII IIII", },
{ "sfleui",  "rA,I",         "0xA 1001 ---- AAAA IIII IIII IIII IIII", },
 
{ "add",     "rD,rA,rB",     "0xC 000D DDDD AAAA BBBB ---- ---- ----", },
{ "sub",     "rD,rA,rB",     "0xC 001D DDDD AAAA BBBB ---- ---- ----", },
{ "and",     "rD,rA,rB",     "0xC 010D DDDD AAAA BBBB ---- ---- ----", },
{ "or",      "rD,rA,rB",     "0xC 100D DDDD AAAA BBBB ---- ---- ----", },
{ "xor",     "rD,rA,rB",     "0xC 110D DDDD AAAA BBBB ---- ---- ----", },
{ "mul",     "rD,rA,rB",     "0xD 000D DDDD AAAA BBBB ---- ---- ----", },
{ "sll",     "rD,rA,rB",     "0xE 000D DDDD AAAA BBBB ---- ---- ----", },
{ "srl",     "rD,rA,rB",     "0xE 100D DDDD AAAA BBBB ---- ---- ----", },
{ "sra",     "rD,rA,rB",     "0xE 101D DDDD AAAA BBBB ---- ---- ----", },

{ "", "", "" }		/* Dummy entry, not included in NUM_OPCODES.  This
			   lets code examine entry i+1 without checking
			   if we've run off the end of the table.  */
};

CONST unsigned int num_opcodes = (((sizeof or1_opcodes) / (sizeof or1_opcodes[0])) - 1);
